# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

license = '''# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

'''

boilerplate = '''
@ffi.def_extern()
def ANARIStatusCallback_python(usrPtr, device, source, sourceType, severity, code, message):
    cb = ffi.from_handle(usrPtr)
    cb(device, source, sourceType, severity, code, ffi.string(message).decode('utf-8'))

'''

conversions = '''
def _from_string_list(x):
    result = []
    if x != ffi.NULL:
        i = 0
        while x[i] != ffi.NULL:
            result.append(str(ffi.string(x[i]), 'utf-8'))
            i += 1
    return result

def _from_type_list(x):
    result = []
    if x != ffi.NULL:
        i = 0
        while x[i] != lib.ANARI_UNKNOWN:
            result.append(x[i])
            i += 1
    return result

def _from_parameter_list(x):
    result = []
    if x != ffi.NULL:
        i = 0
        while x[i].name != ffi.NULL:
            result.append((str(ffi.string(x[i].name), 'utf-8'), x[i].type))
            i += 1
    return result

def _convert_pointer(dataType, x):
    if x == ffi.NULL:
        return None

    x = ffi.cast(_basepointer[dataType], x)

    if dataType == lib.ANARI_STRING_LIST:
        return _from_string_list(x)
    elif dataType == lib.ANARI_DATA_TYPE_LIST:
        return _from_type_list(x)
    elif dataType == lib.ANARI_PARAMETER_LIST:
        return _from_parameter_list(x)
    elif dataType == lib.ANARI_STRING:
        return str(ffi.string(x), 'utf-8')
    elif _elements[dataType]==1:
        return x[0]
    elif _elements[dataType]==2:
        return (x[0], x[1])
    elif _elements[dataType]==3:
        return (x[0], x[1], x[2])
    elif _elements[dataType]==4:
        return (x[0], x[1], x[2], x[3])
    else:
        return ptr

'''

objects = {
    'ANARIDevice',
    'ANARIObject',
    'ANARIArray',
    'ANARIArray1D',
    'ANARIArray2D',
    'ANARIArray3D',
    'ANARICamera',
    'ANARIFrame',
    'ANARIGeometry',
    'ANARIGroup',
    'ANARIInstance',
    'ANARILight',
    'ANARIMaterial',
    'ANARIRenderer',
    'ANARISurface',
    'ANARISampler',
    'ANARISpatialField',
    'ANARIVolume',
    'ANARIWorld'
}

def convert_argument(arg):
    if arg['type'] == 'const char*':
        return arg['name']+'.encode(\'utf-8\')'
    else:
        return arg['name']


def write_wrappers(anari, name):
    code = license
    code += 'from .' + name + ' import lib, ffi\n'
    code += boilerplate

    for enum in anari['enums']:
        for value in enum['values']:
            code += value['name'][6:]+' = lib.'+value['name']+'\n'
        code += '\n'

    types = next((x for x in anari['enums'] if x['name'] == 'ANARIDataType'), None)

    code += '_conversions = {\n'
    for value in types['values']:
        if value['elements'] == 1:
            code += '    %s : \'%s*\',\n'%(value['name'][6:], value['baseType'])
        else:
            code += '    %s : \'%s[%d]\',\n'%(value['name'][6:], value['baseType'], value['elements'])
    code += '}\n'

    code += '_typeof = {\n'
    for value in types['values']:
        if value['elements'] == 1:
            code += '    %s : \'%s\',\n'%(value['name'][6:], value['baseType'])
        else:
            code += '    %s : \'%s[%d]\',\n'%(value['name'][6:], value['baseType'], value['elements'])
    code += '}\n'

    code += '_basepointer = {\n'
    for value in types['values']:
        if value['baseType'][-1] == '*':
            code += '    %s : \'%s\',\n'%(value['name'][6:], value['baseType'])
        else:
            code += '    %s : \'%s*\',\n'%(value['name'][6:], value['baseType'])
    code += '}\n'

    code += '_elements = {\n'
    for value in types['values']:
        code += '    %s : %s,\n'%(value['name'][6:], value['elements'])
    code += '}\n'

    code += conversions

    code += '\n'

    return code



# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

import sys
import os
import json
import hash_gen
import merge_anari
import argparse
import pathlib

class QueryGenerator:
    def __init__(self, anari):
        self.anari = anari
        self.named_objects = dict()
        self.anon_objects = dict()
        self.param_strings = set()
        self.objects = dict()
        type_enums = next(x for x in anari["enums"] if x["name"] == "ANARIDataType")
        self.type_enum_dict = {e["name"]: e for e in type_enums["values"]}

        offset = 0
        for obj in anari["objects"]:
            self.param_strings.update([p["name"] for p in obj["parameters"]])

            if not obj["type"] in self.objects:
                self.objects[obj["type"]] = []

            parameter_list = []
            parameter_type_list = []
            for param in obj["parameters"]:
                parameter_list.append(param)
                for t in param["types"]:
                    parameter_type_list.append((param["name"], t))

            func_name = obj["type"]
            length = len(parameter_list)
            name = 0
            param_data = {"parameters": parameter_list, "range": range(offset, offset+length), "parameters_with_types": parameter_type_list}
            if "name" in obj:
                self.objects[obj["type"]].append(obj["name"])
                self.named_objects[(obj["type"],obj["name"])] = param_data
                func_name += "_"+obj["name"]
                name = "\""+obj["name"]+"\""
            else:
                self.anon_objects[obj["type"]] = param_data

            offset += length

        self.param_strings = sorted(list(self.param_strings))
        self.named_types = sorted(list(set([k[0] for k in self.named_objects])))
        self.subtype_list = sorted(list(set([k[1] for k in self.named_objects])))
        self.attribute_list = [x["name"] for x in anari["attributes"]]
        self.info_strings = ["required", "default", "minimum", "maximum", "description", "elementType", "value", "sourceExtension", "extension", "parameter", "channel", "use"]
        self.uses = {"color", "direction", "point", "vector"}

    def format_as(self, value, anari_type):
        if value is None:
            return "nullptr"
        basetype = self.type_enum_dict[anari_type]["baseType"]
        if isinstance(value, str):
            return value
        elif basetype == "float":
            return "%ff"%value
        elif basetype == "double":
            return "%f"%value
        elif basetype.startswith("uint") or basetype.startswith("int"):
            macro = basetype.upper()[:-1]+'C'
            return "%s(%d)"%(macro,value)
        else:
            return str(value)

    def preamble(self):
        code = "static " + hash_gen.gen_hash_function("subtype_hash", self.subtype_list)
        code += "static " + hash_gen.gen_hash_function("param_hash", self.param_strings)
        code += "static " + hash_gen.gen_hash_function("info_hash", self.info_strings)
        code += "static const int32_t anari_true = 1;\n"
        code += "static const int32_t anari_false = 0;\n"
        return code;

    def generate_extension_query(self):
        code = "const char ** query_extensions() {\n"
        code += "   static const char *extensions[] = {\n      "
        if self.anari["extensions"]:
            code += ",\n      ".join(["\"ANARI_%s\""%x for x in self.anari["extensions"]]) + ",\n"
        code += "      0\n   };\n"
        code += "   return extensions;\n"
        code += "}\n"
        return code

    def generate_subtype_query(self):
        code = "const char ** query_object_types(ANARIDataType type) {\n"
        code += "   switch(type) {\n"
        for key, value in self.objects.items():
            if value:
                code += "      case %s:\n"%(key)
                code += "      {\n"
                code += "         static const char *%s_subtypes[] = {"%(key)
                code += ", ".join(["\"%s\""%x for x in value])+", 0};\n"
                code += "         return %s_subtypes;\n"%key
                code += "      }\n"
        code += "      default:\n"
        code += "      {\n"
        code += "         static const char *none_subtypes[] = {0};\n"
        code += "         return none_subtypes;\n"
        code += "      }\n"
        code += "   }\n"
        code += "}\n"
        return code

    def check_and_return(self, value_type_name, value):
        value_type = self.type_enum_dict[value_type_name]
        code = ""
        code += "         if("
        code += "paramType == " + value_type["name"] + " && "
        code += "infoType == " + value_type["name"] + ") {\n"
        if value_type["name"] == "ANARI_STRING":
            code += "            static const char *default_value = \"%s\";\n"%value
        else:
            code += "            static const " + value_type["baseType"]
            code += " default_value[%d]"%value_type["elements"] + " = "
            if isinstance(value, list):
                code += "{" + ", ".join([self.format_as(x, value_type_name) for x in value]) + "};\n"
            else:
                code += "{" + self.format_as(value, value_type_name) + "};\n"
        code += "            return default_value;\n"
        code += "         } else {\n"
        code += "            return nullptr;\n"
        code += "         }\n"
        return code

    def generate_parameter_info_query(self):
        code = ""
        # generate the per parameter query functions

        for obj in self.anari["objects"]:
            objname = obj["type"]
            if "name" in obj:
                objname += "_" + obj["name"]

            for param in obj["parameters"]:
                paramname = param["name"].replace(".", "_").replace(":", "_")
                code += "static const void * " + objname + "_" + paramname + "_info(ANARIDataType paramType, int infoName, ANARIDataType infoType) {\n"
                code += "   (void)paramType;\n"
                code += "   switch(infoName) {\n"

                # required info always exists
                code += "      case "+str(self.info_strings.index("required"))+": // required\n"
                code += "         if(infoType == ANARI_BOOL) {\n"
                if "required" in param["tags"]:
                    code += "            return &anari_true;\n"
                else:
                    code += "            return &anari_false;\n"
                code += "         } else {\n"
                code += "            return nullptr;\n"
                code += "         }\n"

                if "default" in param:
                    code += "      case "+str(self.info_strings.index("default"))+": // default\n"
                    code += self.check_and_return(param["types"][0], param["default"])

                use = self.uses.intersection(param["tags"])
                if use:
                    code += "      case "+str(self.info_strings.index("use"))+": // use\n"
                    code += "         if(infoType == ANARI_STRING) {\n"
                    code += "            return \""+use.pop()+"\";\n"
                    code += "         } else {\n"
                    code += "            return nullptr;\n"
                    code += "         }\n"

                if "minimum" in param:
                    code += "      case "+str(self.info_strings.index("minimum"))+": // minimum\n"
                    code += self.check_and_return(param["types"][0], param["minimum"])

                if "maximum" in param:
                    code += "      case "+str(self.info_strings.index("maximum"))+": // maximum\n"
                    code += self.check_and_return(param["types"][0], param["maximum"])

                if "description" in param:
                    code += "      case "+str(self.info_strings.index("description"))+": // description\n"
                    code += "         {\n"
                    code += "            static const char *description = \"%s\";\n"%param["description"]
                    code += "            return description;\n"
                    code += "         }\n"

                if "elementType" in param:
                    code += "      case "+str(self.info_strings.index("elementType"))+": // elementType\n"
                    code += "         if(infoType == ANARI_DATA_TYPE_LIST) {\n"
                    code += "            static const ANARIDataType values[] = {"
                    code += ", ".join(param["elementType"]) + ", ANARI_UNKNOWN};\n"
                    code += "            return values;\n"
                    code += "         } else {\n"
                    code += "            return nullptr;\n"
                    code += "         }\n"

                if "values" in param:
                    code += "      case "+str(self.info_strings.index("value"))+": // value\n"
                    if param["types"][0] == "ANARI_STRING":
                        code += "         if(paramType == ANARI_STRING && infoType == ANARI_STRING_LIST) {\n"
                        code += "            static const char *values[] = {"
                        code += ", ".join(["\"%s\""%v for v in param["values"]]) + ", nullptr};\n"
                        code += "            return values;\n"
                    elif param["types"][0] == "ANARI_DATA_TYPE":
                        code += "         if(paramType == ANARI_DATA_TYPE && infoType == ANARI_DATA_TYPE_LIST) {\n"
                        code += "            static const ANARIDataType values[] = {"
                        code += ", ".join(param["values"]) + ", ANARI_UNKNOWN};\n"
                        code += "            return values;\n"
                    code += "         } else {\n"
                    code += "            return nullptr;\n"
                    code += "         }\n"

                if "attribute" in param["tags"]:
                    code += "      case "+str(self.info_strings.index("value"))+": // value\n"
                    code += "         if(paramType == ANARI_STRING && infoType == ANARI_STRING_LIST) {\n"
                    code += "            static const char *values[] = {"
                    code += ", ".join(["\"%s\""%v for v in self.attribute_list]) + ", nullptr};\n"
                    code += "            return values;\n"
                    code += "         } else {\n"
                    code += "            return nullptr;\n"
                    code += "         }\n"

                if "sourceExtension" in param:
                    code += "      case "+str(self.info_strings.index("sourceExtension"))+": // sourceExtension\n"
                    code += "         if(infoType == ANARI_STRING) {\n"
                    code += "            static const char *extension = \"%s\";\n"%param["sourceExtension"]
                    code += "            return extension;\n"
                    code += "         } else if(infoType == ANARI_INT32) {\n"
                    code += "            static const int32_t value = %d;\n"%self.anari["extensions"].index(param["sourceExtension"])
                    code += "            return &value;\n"
                    code += "         }\n"

                code += "      default: return nullptr;\n"
                code += "   }\n"
                code += "}\n"


            code += "static const void * " + objname + "_param_info(const char *paramName, ANARIDataType paramType, int infoName, ANARIDataType infoType) {\n"
            code += "   switch(param_hash(paramName)) {\n"
            for param in obj["parameters"]:
                paramname = param["name"].replace(".", "_").replace(":", "_")
                code += "      case %i:\n"%self.param_strings.index(param["name"])
                code += "         return " + objname + "_" + paramname + "_info(paramType, infoName, infoType);\n"
            code += "      default:\n"
            code += "         return nullptr;\n"
            code += "   }\n"
            code += "}\n"




        for type_enum in self.named_types:
            subtypes = [key[1] for key,params in self.named_objects.items() if key[0] == type_enum]
            code += "static const void * "+type_enum+"_param_info(const char *subtype, const char *paramName, ANARIDataType paramType, int infoName, ANARIDataType infoType) {\n"
            code += "   switch(subtype_hash(subtype)) {\n"
            for subtype in subtypes:
                code += "      case %d:\n"%(self.subtype_list.index(subtype))
                code += "         return %s_%s_param_info(paramName, paramType, infoName, infoType);\n"%(type_enum, subtype)
            code += "      default:\n"
            code += "         return nullptr;\n"
            code += "   }\n"
            code += "}\n"

        code += "const void * query_param_info_enum(ANARIDataType type, const char *subtype, const char *paramName, ANARIDataType paramType, int infoName, ANARIDataType infoType) {\n"
        code += "   switch(type) {\n"
        for type_enum in self.named_types:
            code += "      case %s:\n"%(type_enum)
            code += "         return %s_param_info(subtype, paramName, paramType, infoName, infoType);\n"%type_enum

        for type_enum in self.anon_objects.keys():
            code += "      case %s:\n"%(type_enum)
            code += "         return %s_param_info(paramName, paramType, infoName, infoType);\n"%type_enum

        code += "      default:\n"
        code += "         return nullptr;\n"
        code += "   }\n"
        code += "}\n"

        code += "const void * query_param_info(ANARIDataType type, const char *subtype, const char *paramName, ANARIDataType paramType, const char *infoNameString, ANARIDataType infoType) {\n"
        code += "   int infoName = info_hash(infoNameString);\n"
        code += "   return query_param_info_enum(type, subtype, paramName, paramType, infoName, infoType);\n"
        code += "}\n"

        return code

    def generate_object_info_query(self):
        code = ""
        for obj in self.anari["objects"]:
            objname = obj["type"]
            if "name" in obj:
                objname += "_" + obj["name"]

            code += "static const void * " + objname + "_info(int infoName, ANARIDataType infoType) {\n"
            code += "   switch(infoName) {\n"
            if "description" in obj:
                code += "      case "+str(self.info_strings.index("description"))+": // description\n"
                code += "         {\n"
                code += "            static const char *description = \"%s\";\n"%obj["description"]
                code += "            return description;\n"
                code += "         }\n"

            if "parameters" in obj:
                code += "      case "+str(self.info_strings.index("parameter"))+": // parameter\n"
                code += "         if(infoType == ANARI_PARAMETER_LIST) {\n"
                code += "            static const ANARIParameter parameters[] = {\n"
                for param in obj["parameters"]:
                    for t in param["types"]:
                        code += "               {\"%s\", %s},\n"%(param["name"], t)
                code += "               {0, ANARI_UNKNOWN}\n"
                code += "            };\n"
                code += "            return parameters;\n"
                code += "         } else {\n"
                code += "            return nullptr;\n"
                code += "         }\n"

            if "sourceExtension" in obj:
                code += "      case "+str(self.info_strings.index("sourceExtension"))+": // sourceExtension\n"
                code += "         if(infoType == ANARI_STRING) {\n"
                code += "            static const char *extension = \"%s\";\n"%obj["sourceExtension"]
                code += "            return extension;\n"
                code += "         } else if(infoType == ANARI_INT32) {\n"
                code += "            static const int value = %d;\n"%self.anari["extensions"].index(obj["sourceExtension"])
                code += "            return &value;\n"
                code += "         } else {\n"
                code += "            return nullptr;\n"
                code += "         }\n"
            if "channel" in obj:
                code += "      case "+str(self.info_strings.index("channel"))+": // channel\n"
                code += "         if(infoType == ANARI_STRING_LIST) {\n"
                code += "            static const char *channel[] = {\n"
                code += "               " + ",\n               ".join(["\"%s\""%f for f in obj["channel"]])+",\n"
                code += "               0\n"
                code += "            };\n"
                code += "            return channel;\n"
                code += "         } else {\n"
                code += "            return nullptr;\n"
                code += "         }\n"

            if "extension" in obj:
                code += "      case "+str(self.info_strings.index("extension"))+": // extension\n"
                code += "         if(infoType == ANARI_STRING_LIST) {\n"
                code += "            static const char *extensions[] = {\n"
                code += "               " + ",\n               ".join(["\"ANARI_%s\""%f for f in obj["extension"]])+",\n"
                code += "               0\n"
                code += "            };\n"
                code += "            return extensions;\n"
                code += "         } else {\n"
                code += "            return nullptr;\n"
                code += "         }\n"
            elif obj["type"] == "ANARI_DEVICE" or obj["type"] == "ANARI_RENDERER":
                code += "      case "+str(self.info_strings.index("extension"))+": // extension\n"
                code += "         if(infoType == ANARI_STRING_LIST) {\n"
                code += "            static const char *extensions[] = {\n"
                code += "               " + ",\n               ".join(["\"ANARI_%s\""%f for f in self.anari["extensions"]])+",\n"
                code += "               0\n"
                code += "            };\n"
                code += "            return extensions;\n"
                code += "         } else {\n"
                code += "            return nullptr;\n"
                code += "         }\n"


            code += "      default: return nullptr;\n"
            code += "   }\n"
            code += "}\n"



        for type_enum in self.named_types:
            subtypes = [key[1] for key,params in self.named_objects.items() if key[0] == type_enum]
            code += "static const void * "+type_enum+"_info(const char *subtype, int infoName, ANARIDataType infoType) {\n"
            code += "   switch(subtype_hash(subtype)) {\n"
            for subtype in subtypes:
                code += "      case %d:\n"%(self.subtype_list.index(subtype))
                code += "         return %s_%s_info(infoName, infoType);\n"%(type_enum, subtype)
            code += "      default:\n"
            code += "         return nullptr;\n"
            code += "   }\n"
            code += "}\n"

        code += "const void * query_object_info_enum(ANARIDataType type, const char *subtype, int infoName, ANARIDataType infoType) {\n"
        code += "   switch(type) {\n"
        for type_enum in self.named_types:
            code += "      case %s:\n"%(type_enum)
            code += "         return %s_info(subtype, infoName, infoType);\n"%type_enum

        for type_enum in self.anon_objects.keys():
            code += "      case %s:\n"%(type_enum)
            code += "         return %s_info(infoName, infoType);\n"%type_enum

        code += "      default:\n"
        code += "         return nullptr;\n"
        code += "   }\n"
        code += "}\n"

        code += "const void * query_object_info(ANARIDataType type, const char *subtype, const char *infoNameString, ANARIDataType infoType) {\n"
        code += "   int infoName = info_hash(infoNameString);"
        code += "   return query_object_info_enum(type, subtype, infoName, infoType);"
        code += "}\n"

        return code

    def generate_query_declarations(self):
        code = ""
        for idx, info in enumerate(self.info_strings):
            code += "#define ANARI_INFO_%s %d\n"%(info, idx)
        code += "const int extension_count = %d;\n"%len(self.anari["extensions"])
        code += "const char ** query_extensions();\n"
        code += "const char ** query_object_types(ANARIDataType type);\n"
        code += "const ANARIParameter * query_params(ANARIDataType type, const char *subtype);\n"
        code += "const void * query_param_info_enum(ANARIDataType type, const char *subtype, const char *paramName, ANARIDataType paramType, int infoName, ANARIDataType infoType);\n"
        code += "const void * query_param_info(ANARIDataType type, const char *subtype, const char *paramName, ANARIDataType paramType, const char *infoNameString, ANARIDataType infoType);\n"
        code += "const void * query_object_info_enum(ANARIDataType type, const char *subtype, int infoName, ANARIDataType infoType);\n"
        code += "const void * query_object_info(ANARIDataType type, const char *subtype, const char *infoNameString, ANARIDataType infoType);\n"
        return code

parser = argparse.ArgumentParser(description="Generate query functions for an ANARI device.")
parser.add_argument("-d", "--device", dest="devicespec", type=open, help="The device json file.")
parser.add_argument("-j", "--json", dest="json", type=pathlib.Path, action="append", help="Path to the core and extension json root.")
parser.add_argument("-p", "--prefix", dest="prefix", help="Prefix for the classes and filenames.")
parser.add_argument("-n", "--namespace", dest="namespace", help="Namespace for the classes and filenames.")
parser.add_argument("-o", "--output", dest="outdir", type=pathlib.Path, default=pathlib.Path("."), help="Output directory")
args = parser.parse_args()


#flattened list of all input jsons in supplied directories
jsons = [entry for j in args.json for entry in j.glob("**/*.json")]

#load the device root
device = json.load(args.devicespec)
merge_anari.tag_extension(device)
print("opened " + device["info"]["type"] + " " + device["info"]["name"])

dependencies = merge_anari.crawl_dependencies(device, jsons)
#merge all dependencies
for x in dependencies:
    matches = [p for p in jsons if p.stem == x]
    for m in matches:
        extension = json.load(open(m))
        merge_anari.assemble(extension, jsons)
        merge_anari.tag_extension(extension)
        merge_anari.merge(device, extension)

#generate files
gen = QueryGenerator(device)


def begin_namespaces(args):
    output = ""
    if args.namespace:
        for n in args.namespace.split("::"):
            output += "namespace %s {\n"%n
    return output

def end_namespaces(args):
    output = ""
    if args.namespace:
        for n in args.namespace.split("::"):
            output += "}\n"
    return output

with open(args.outdir/(args.prefix + "_queries.cpp"), mode='w') as f:
    f.write("// Copyright 2024-2026 The Khronos Group\n")
    f.write("// SPDX-License-Identifier: Apache-2.0\n\n")
    f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
    f.write("// Don't make changes to this directly\n\n")

    f.write("#include <stdint.h>\n")
    f.write("#include <math.h>\n")
    f.write("#include <anari/anari.h>\n")
    f.write(begin_namespaces(args))
    f.write(gen.preamble())
    f.write(gen.generate_extension_query())
    f.write(gen.generate_subtype_query())
    f.write(gen.generate_parameter_info_query())
    f.write(gen.generate_object_info_query())
    f.write(end_namespaces(args))


with open(args.outdir/(args.prefix + "_queries.h"), mode='w') as f:
    f.write("// Copyright 2024-2026 The Khronos Group\n")
    f.write("// SPDX-License-Identifier: Apache-2.0\n\n")
    f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
    f.write("// Don't make changes to this directly\n\n")

    f.write("#include <anari/anari.h>\n")
    f.write(begin_namespaces(args))
    f.write(gen.generate_query_declarations())
    f.write(end_namespaces(args))



# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

from cffi import FFI
import sys
import os
import json
import merge_anari
import argparse
import pathlib
import glob

import device_wrapper


parser = argparse.ArgumentParser(description="generate cffi bindings")

parser.add_argument("-n", "--name", dest="name", type=str, help="module name")
parser.add_argument("-d", "--device", dest="devicespec", type=open, help="The device json file.")
parser.add_argument("-j", "--json", dest="json", type=pathlib.Path, action="append", help="Path to the core and extension json root.")
parser.add_argument("-i", "--include", dest="includes", type=str, action="append", help="include")
parser.add_argument("-R", "--forcereleaselink", action="store_true", help="Force the extension to implicitly link to a release python library on Windows, overriding possible debug python linking")
parser.add_argument("-O", "--output", dest="outdir", type=pathlib.Path, help="Output directory")
args = parser.parse_args()

#flattened list of all input jsons in supplied directories
jsons = [entry for j in args.json for entry in j.glob("**/*.json")]

#load the device root
device = json.load(args.devicespec)
merge_anari.tag_extension(device)
print("opened " + device["info"]["type"] + " " + device["info"]["name"])

dependencies = merge_anari.crawl_dependencies(device, jsons)
#merge all dependencies
for x in dependencies:
    matches = [p for p in jsons if p.stem == x]
    for m in matches:
        extension = json.load(open(m))
        merge_anari.tag_extension(extension)
        merge_anari.merge(device, extension)

def write_definitions(anari):
    defs = ''

    # generate enums
    for enum in anari['enums']:
        defs += 'typedef '+enum['baseType']+' '+enum['name']+';\n'
        for value in enum['values']:
            defs += '#define '+value['name']+' '+str(value['value'])+'\n'
        defs += '\n'

    for opaque in anari['opaqueTypes']:
        defs += 'typedef void* '+opaque['name']+';\n'

    defs += '\n'

    # generate structs
    for struct in anari['structs']:
        defs += 'typedef struct {\n'

        if struct['members']:
            for member in struct['members']:
                if 'elements' in member:
                    defs += '  '+member['type']+' '+member['name']+'['+str(member['elements'])+'];\n'
                else:
                    defs += '  '+member['type']+' '+member['name']+';\n'

        defs += '} '+struct['name']+';\n'

    defs += '\n'

    for fun in anari['functionTypedefs']:
        defs += 'typedef '+fun['returnType']+' (*'+fun['name']+')('

        if fun['arguments']:
            arg = fun['arguments'][0]
            defs += arg['type']+' '+arg['name']
            for arg in fun['arguments'][1:]:
                defs += ', '+arg['type']+' '+arg['name']

        defs += ');\n'

    defs += '\n'


    for fun in anari['functions']:
        defs += fun['returnType']+' '+fun['name']+'('

        if fun['arguments']:
            arg = fun['arguments'][0]
            defs += arg['type']+' '+arg['name']
            for arg in fun['arguments'][1:]:
                defs += ', '+arg['type']+' '+arg['name']

        defs += ');\n'


    defs += '\n'


    for fun in anari['functionTypedefs']:
        defs += 'extern "Python" ' + fun['returnType']+' '+fun['name']+'_python('

        if fun['arguments']:
            arg = fun['arguments'][0]
            defs += arg['type']+' '+arg['name']
            for arg in fun['arguments'][1:]:
                defs += ', '+arg['type']+' '+arg['name']

        defs += ');\n'


    return defs

defs = write_definitions(device)

# print(defs)

includes = '#include "anari/anari.h"\n'
if args.includes:
    for i in args.includes:
        includes += '#include "'+i+'"\n'

ffibuilder = FFI()
ffibuilder.cdef(defs)
ffibuilder.set_source(args.name, includes)

bindings_file = args.outdir/'anari_api_bindings.c'
ffibuilder.emit_c_code(str(bindings_file))
if args.forcereleaselink:
    # Add an undef _DEBUG to the file to force compilation in release mode
    with open(bindings_file, 'r') as file:
        existing_content = file.read()
    with open(bindings_file, 'w') as file:
        file.write('#undef _DEBUG' + '\n' + existing_content)

wrapper = device_wrapper.write_wrappers(device, args.name)

with open(args.outdir/'utils.py', mode='w') as f:
    f.write(wrapper)


# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

from cffi import FFI
import sys
import os
import json
import merge_anari
import argparse
import pathlib
import glob


parser = argparse.ArgumentParser(description="Generate query functions for an ANARI device.")
parser.add_argument("-d", "--device", dest="devicespec", type=open, help="The device json file.")
parser.add_argument("-j", "--json", dest="json", type=pathlib.Path, action="append", help="Path to the core and extension json root.")
parser.add_argument("-I", "--include", dest="incdirs", type=pathlib.Path, action="append", help="Include directories")
parser.add_argument("-L", "--lib", dest="libdirs", type=pathlib.Path, action="append", help="Library directories")
parser.add_argument("-R", "--forcereleaselink", action="store_true", help="Force the extension to implicitly link to a release python library on Windows, overriding possible debug python linking")
args = parser.parse_args()

#flattened list of all input jsons in supplied directories
jsons = [entry for j in args.json for entry in j.glob("**/*.json")]

#load the device root
device = json.load(args.devicespec)
merge_anari.tag_extension(device)
print("opened " + device["info"]["type"] + " " + device["info"]["name"])

dependencies = merge_anari.crawl_dependencies(device, jsons)
#merge all dependencies
for x in dependencies:
    matches = [p for p in jsons if p.stem == x]
    for m in matches:
        extension = json.load(open(m))
        merge_anari.tag_extension(extension)
        merge_anari.merge(device, extension)

def write_definitions(anari):
    defs = ''

    # generate enums
    for enum in anari['enums']:
        defs += 'typedef '+enum['baseType']+' '+enum['name']+';\n'
        for value in enum['values']:
            defs += '#define '+value['name']+' '+str(value['value'])+'\n'
        defs += '\n'

    for opaque in anari['opaqueTypes']:
        defs += 'typedef void* '+opaque['name']+';\n'

    defs += '\n'

    # generate structs
    for struct in anari['structs']:
        defs += 'typedef struct {\n'

        if struct['members']:
            for member in struct['members']:
                if 'elements' in member:
                    defs += '  '+member['type']+' '+member['name']+'['+str(member['elements'])+'];\n'
                else:
                    defs += '  '+member['type']+' '+member['name']+';\n'

        defs += '} '+struct['name']+';\n'

    defs += '\n'

    for fun in anari['functionTypedefs']:
        defs += 'typedef '+fun['returnType']+' (*'+fun['name']+')('

        if fun['arguments']:
            arg = fun['arguments'][0]
            defs += arg['type']+' '+arg['name']
            for arg in fun['arguments'][1:]:
                defs += ', '+arg['type']+' '+arg['name']

        defs += ');\n'

    defs += '\n'


    for fun in anari['functions']:
        defs += fun['returnType']+' '+fun['name']+'('

        if fun['arguments']:
            arg = fun['arguments'][0]
            defs += arg['type']+' '+arg['name']
            for arg in fun['arguments'][1:]:
                defs += ', '+arg['type']+' '+arg['name']

        defs += ');\n'


    defs += '\n'


    for fun in anari['functionTypedefs']:
        defs += 'extern "Python" ' + fun['returnType']+' '+fun['name']+'_python('

        if fun['arguments']:
            arg = fun['arguments'][0]
            defs += arg['type']+' '+arg['name']
            for arg in fun['arguments'][1:]:
                defs += ', '+arg['type']+' '+arg['name']

        defs += ');\n'


    return defs

defs = write_definitions(device)

ffibuilder = FFI()
ffibuilder.cdef(defs)
ffibuilder.set_source('pyanari',
'''
#include "anari/anari.h"
'''
#    , libraries = ['anari'],
#    library_dirs = [str(x) for x in args.libdirs],
#    include_dirs = [str(x) for x in args.incdirs]
    )

if __name__ == "__main__":
    bindings_file = 'pyanari_bindings.c'
    ffibuilder.emit_c_code(bindings_file)
    if args.forcereleaselink:
        # Add an undef _DEBUG to the file to force compilation in release mode
        with open(bindings_file, 'r') as file:
            existing_content = file.read()
        with open(bindings_file, 'w') as file:
            file.write('#undef _DEBUG' + '\n' + existing_content)
    #ffibuilder.compile(verbose=True, debug=False)
    #this ends up with a funny python version specific name despite being version agnostic
    #pyanari_files = glob.glob('pyanari.*.so') + glob.glob('pyanari.*.pyd')
    #for pyanari_file in pyanari_files:
    #    fileWithExt = os.path.splitext(pyanari_file)
    #    newFileName = f'pyanari{fileWithExt[1]}'
    #    print(f'Renamed {pyanari_file} to {newFileName}')
    #    os.replace(pyanari_file, newFileName)


boilerplate = '''# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

from pyanari import lib, ffi

@ffi.def_extern()
def ANARIStatusCallback_python(usrPtr, device, source, sourceType, severity, code, message):
    cb = ffi.from_handle(usrPtr)
    cb(device, source, sourceType, severity, code, ffi.string(message).decode('utf-8'))

def _releaseBoth(device, handle):
    lib.anariRelease(device, handle)
    lib.anariRelease(device, device)

'''

conversions = '''
def _from_string_list(x):
    result = []
    if x != ffi.NULL:
        i = 0
        while x[i] != ffi.NULL:
            result.append(str(ffi.string(x[i]), 'utf-8'))
            i += 1
    return result

def _from_type_list(x):
    result = []
    if x != ffi.NULL:
        i = 0
        while x[i] != lib.ANARI_UNKNOWN:
            result.append(x[i])
            i += 1
    return result

def _from_parameter_list(x):
    result = []
    if x != ffi.NULL:
        i = 0
        while x[i].name != ffi.NULL:
            result.append((str(ffi.string(x[i].name), 'utf-8'), x[i].type))
            i += 1
    return result

def _convert_pointer(dataType, x):
    if x == ffi.NULL:
        return None

    x = ffi.cast(_basepointer[dataType], x)

    if dataType == lib.ANARI_STRING_LIST:
        return _from_string_list(x)
    elif dataType == lib.ANARI_DATA_TYPE_LIST:
        return _from_type_list(x)
    elif dataType == lib.ANARI_PARAMETER_LIST:
        return _from_parameter_list(x)
    elif dataType == lib.ANARI_STRING:
        return str(ffi.string(x), 'utf-8')
    elif _elements[dataType]==1:
        return x[0]
    elif _elements[dataType]==2:
        return (x[0], x[1])
    elif _elements[dataType]==3:
        return (x[0], x[1], x[2])
    elif _elements[dataType]==4:
        return (x[0], x[1], x[2], x[3])
    else:
        return ptr

'''


special = {
    'anariSetParameter' :
    '''def anariSetParameter(device, handle, name, atype, value):
    if isinstance(value, str):
        value = value.encode('utf-8')
        lib.anariSetParameter(
            device, handle, name.encode('utf-8'),
            atype, ffi.new('const char[]', value))
    else:
        lib.anariSetParameter(
            device, handle, name.encode('utf-8'),
            atype, ffi.new(_conversions[atype], value))

''',
    'anariLoadLibrary' :
    '''def anariLoadLibrary(name, callback = ffi.NULL):
    cbwrapper = ffi.NULL if callback == ffi.NULL else lib.ANARIStatusCallback_python
    device = lib.anariLoadLibrary(name.encode('utf-8'), cbwrapper, callback)
    if device == ffi.NULL:
        return None
    else:
        return ffi.gc(device, lib.anariUnloadLibrary)

''',
    'anariNewArray1D' :
    '''def anariNewArray1D(device, appMemory, dataType, numItems1):
    result = lib.anariNewArray1D(device, ffi.NULL, ffi.NULL, ffi.NULL, dataType, numItems1)
    ptr = lib.anariMapArray(device, result)
    ffi.memmove(ptr, appMemory, ffi.sizeof(_typeof[dataType])*numItems1)
    lib.anariUnmapArray(device, result)
    lib.anariRetain(device, device)
    return ffi.gc(result, lambda h, d=device: _releaseBoth(d, h))

''',
    'anariNewArray2D' :
    '''def anariNewArray2D(device, appMemory, dataType, numItems1, numItems2):
    result = lib.anariNewArray2D(device, ffi.NULL, ffi.NULL, ffi.NULL, dataType, numItems1, numItems2)
    ptr = lib.anariMapArray(device, result)
    ffi.memmove(ptr, appMemory, ffi.sizeof(_typeof[dataType])*numItems1*numItems2)
    lib.anariUnmapArray(device, result)
    lib.anariRetain(device, device)
    return ffi.gc(result, lambda h, d=device: _releaseBoth(d, h))

''',
    'anariNewArray3D' :
    '''def anariNewArray3D(device, appMemory, dataType, numItems1, numItems2, numItems3):
    result = lib.anariNewArray3D(device, ffi.NULL, ffi.NULL, ffi.NULL, dataType, numItems1, numItems2, numItems3)
    ptr = lib.anariMapArray(device, result)
    ffi.memmove(ptr, appMemory, ffi.sizeof(_typeof[dataType])*numItems1*numItems2*numItems3)
    lib.anariUnmapArray(device, result)
    lib.anariRetain(device, device)
    return ffi.gc(result, lambda h, d=device: _releaseBoth(d, h))

''',
    'anariMapFrame' :
    '''def anariMapFrame(device, frame, channel):
    frame_width = ffi.new('uint32_t*', 0)
    frame_height = ffi.new('uint32_t*', 0)
    frame_type = ffi.new('ANARIDataType*', 0)
    result = lib.anariMapFrame(device, frame, channel.encode('utf-8'), frame_width, frame_height, frame_type)
    return (result, int(frame_width[0]), int(frame_height[0]), int(frame_type[0]))

''',
    'anariMapParameterArray1D' :
    '''def anariMapParameterArray1D(device, object, name, dataType, numElements1):
    elementStride = ffi.new('uint64_t*', 0)
    result = lib.anariMapParameterArray1D(device, object, name.encode('utf-8'), dataType, numElements1, elementStride)
    result = ffi.cast(_basepointer[dataType], result)
    return (result, int(elementStride[0]))

''',
    'anariMapParameterArray2D' :
    '''def anariMapParameterArray2D(device, object, name, dataType, numElements1, numElements2):
    elementStride = ffi.new('uint64_t*', 0)
    result = lib.anariMapParameterArray2D(device, object, name.encode('utf-8'), dataType, numElements1, numElements2, elementStride)
    result = ffi.cast(_basepointer[dataType], result)
    return (result, int(elementStride[0]))

''',
    'anariMapParameterArray3D' :
    '''def anariMapParameterArray3D(device, object, name, dataType, numElements1, numElements2, numElements3):
    elementStride = ffi.new('uint64_t*', 0)
    result = lib.anariMapParameterArray3D(device, object, name.encode('utf-8'), dataType, numElements1, numElements2, numElements3, elementStride)
    result = ffi.cast(_basepointer[dataType], result)
    return (result, int(elementStride[0]))

''',
    'anariGetDeviceSubtypes' :
    '''def anariGetDeviceSubtypes(library):
    result = lib.anariGetDeviceSubtypes(library)
    return _from_string_list(result)

''',
    'anariGetDeviceExtensions' :
    '''def anariGetDeviceExtensions(library, deviceSubtype):
    result = lib.anariGetDeviceExtensions(library, deviceSubtype.encode('utf-8'))
    return _from_string_list(result)

''',
    'anariGetObjectSubtypes' :
    '''def anariGetObjectSubtypes(device, objectType):
    result = lib.anariGetObjectSubtypes(device, objectType)
    return _from_string_list(result)

''', 'anariGetObjectInfo' :
    '''def anariGetObjectInfo(device, objectType, objectSubtype, infoName, infoType):
    result = lib.anariGetObjectInfo(device, objectType, objectSubtype.encode('utf-8'), infoName.encode('utf-8'), infoType)
    return _convert_pointer(infoType, result)

''', 'anariGetParameterInfo' :
    '''def anariGetParameterInfo(device, objectType, objectSubtype, parameterName, parameterType, infoName, infoType):
    result = lib.anariGetParameterInfo(device, objectType, objectSubtype.encode('utf-8'), parameterName.encode('utf-8'), parameterType, infoName.encode('utf-8'), infoType)
    return _convert_pointer(infoType, result)

''', 'anariGetProperty' :
    '''def anariGetProperty(device, object, name, type, mask):
    if type == ANARI_STRING:
        size = anariGetProperty(device, object, name+'.size', ANARI_UINT64, mask)
        if size is None:
            return None
    else:
        size = ffi.sizeof(_typeof[type])

    mem = ffi.new('char[]', size)
    result = lib.anariGetProperty(device, object, name.encode('utf-8'), type, mem, size, mask)
    if result:
        return _convert_pointer(type, mem)
    else:
        return None

''',


    'anariRelease' : '', # remove these to avoid confusion
    'anariRetain' : '',
    'anariUnloadLibrary' : ''
}

objects = {
    'ANARIDevice',
    'ANARIObject',
    'ANARIArray',
    'ANARIArray1D',
    'ANARIArray2D',
    'ANARIArray3D',
    'ANARICamera',
    'ANARIFrame',
    'ANARIGeometry',
    'ANARIGroup',
    'ANARIInstance',
    'ANARILight',
    'ANARIMaterial',
    'ANARIRenderer',
    'ANARISurface',
    'ANARISampler',
    'ANARISpatialField',
    'ANARIVolume',
    'ANARIWorld'
}

def convert_argument(arg):
    if arg['type'] == 'const char*':
        return arg['name']+'.encode(\'utf-8\')'
    else:
        return arg['name']


def write_wrappers(anari):
    code = boilerplate

    for enum in anari['enums']:
        for value in enum['values']:
            code += value['name']+' = lib.'+value['name']+'\n'
        code += '\n'

    types = next((x for x in anari['enums'] if x['name'] == 'ANARIDataType'), None)

    code += '_conversions = {\n'
    for value in types['values']:
        if value['elements'] == 1:
            code += '    %s : \'%s*\',\n'%(value['name'], value['baseType'])
        else:
            code += '    %s : \'%s[%d]\',\n'%(value['name'], value['baseType'], value['elements'])
    code += '}\n'

    code += '_typeof = {\n'
    for value in types['values']:
        if value['elements'] == 1:
            code += '    %s : \'%s\',\n'%(value['name'], value['baseType'])
        else:
            code += '    %s : \'%s[%d]\',\n'%(value['name'], value['baseType'], value['elements'])
    code += '}\n'

    code += '_basepointer = {\n'
    for value in types['values']:
        if value['baseType'][-1] == '*':
            code += '    %s : \'%s\',\n'%(value['name'], value['baseType'])
        else:
            code += '    %s : \'%s*\',\n'%(value['name'], value['baseType'])
    code += '}\n'

    code += '_elements = {\n'
    for value in types['values']:
        code += '    %s : %s,\n'%(value['name'], value['elements'])
    code += '}\n'

    code += conversions

    for fun in anari['functions']:
        if fun['name'] in special:
            code += special[fun['name']]
            continue

        code += 'def '+fun['name']+'('
        if fun['arguments']:
            code += ', '.join([x['name'] for x in fun['arguments']])
        code += '):\n'
        code += '    '

        if fun['returnType'] != 'void':
            code += 'result = '

        code += 'lib.'+fun['name']+'('
        code += ', '.join([convert_argument(x) for x in fun['arguments']])
        code += ')\n'

        if fun['returnType'] == 'ANARIDevice':
            code += '    return ffi.gc(result, lambda d: lib.anariRelease(d, d))\n'
        elif fun['returnType'] in objects:
            code += '    lib.anariRetain(device, device)\n'
            code += '    return ffi.gc(result, lambda h, d=device: _releaseBoth(d, h))\n'
        elif fun['returnType'] != 'void':
            code += '    return result\n'

        code += '\n'

    code += '\n'

    return code

with open('anari.py', mode='w') as f:
    f.write(write_wrappers(device))



# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

import sys
import os
import json

import merge_anari

def write_anari_enums(filename, anari):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, mode='w') as f:

        f.write("// Copyright 2024-2026 The Khronos Group\n")
        f.write("// SPDX-License-Identifier: Apache-2.0\n\n")
        f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
        f.write("// Don't make changes to this directly\n\n")

        f.write("""
#pragma once
#ifdef __cplusplus
struct ANARIDataType
{
  ANARIDataType() = default;
  constexpr ANARIDataType(int v) noexcept : value(v) {}
  constexpr operator int() const noexcept { return value; }
 private:
  int value;
};
constexpr bool operator<(ANARIDataType v1, ANARIDataType v2)
{
  return static_cast<int>(v1) < static_cast<int>(v2);
}
#define ANARI_DATA_TYPE_DEFINE(v) ANARIDataType(v)
#else
typedef int ANARIDataType;
#define ANARI_DATA_TYPE_DEFINE(v) v
#endif
""")

        # generate enums
        for enum in anari['enums']:
            if enum['name'] != 'ANARIDataType':
                f.write('\ntypedef '+enum['baseType']+' '+enum['name']+';\n')
            for value in enum['values']:
                if enum['name'] == 'ANARIDataType':
                    f.write('#define '+value['name']+' ANARI_DATA_TYPE_DEFINE('+str(value['value'])+')\n')
                else:
                    f.write('#define '+value['name']+' '+str(value['value'])+'\n')

def write_anari_header(filename, anari):
    os.makedirs(os.path.dirname(filename), exist_ok=True)

    with open(filename, mode='w') as f:
        f.write("// Copyright 2021-2026 The Khronos Group\n")
        f.write("// SPDX-License-Identifier: Apache-2.0\n\n")
        f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
        f.write("// Don't make changes to this directly\n\n")

        f.write("""
#pragma once

#include <stdint.h>
#include <sys/types.h>

#ifndef NULL
#if __cplusplus >= 201103L
#define NULL nullptr
#else
#define NULL 0
#endif
#endif

#define ANARI_INVALID_HANDLE NULL

#include "frontend/anari_sdk_version.h"
#include "frontend/anari_enums.h"

#ifdef _WIN32
#ifdef ANARI_STATIC_DEFINE
#define ANARI_INTERFACE
#else
#ifdef anari_EXPORTS
#define ANARI_INTERFACE __declspec(dllexport)
#else
#define ANARI_INTERFACE __declspec(dllimport)
#endif
#endif
#elif defined __GNUC__
#define ANARI_INTERFACE __attribute__((__visibility__("default")))
#else
#define ANARI_INTERFACE
#endif

#ifdef __GNUC__
#define ANARI_DEPRECATED __attribute__((deprecated))
#elif defined(_MSC_VER)
#define ANARI_DEPRECATED __declspec(deprecated)
#else
#define ANARI_DEPRECATED
#endif

#ifdef __cplusplus
// C++ DOES support default initializers
#define ANARI_DEFAULT_VAL(a) = a
#else
/* C99 does NOT support default initializers, so we use this macro
   to define them away */
#define ANARI_DEFAULT_VAL(a)
#endif
""")

        # generate opaque types
        f.write('#ifdef __cplusplus\n')
        f.write('namespace anari {\n')
        f.write('namespace api {\n')

        for opaque in anari['opaqueTypes']:
            if 'parent' in opaque:
                f.write('struct '+opaque['name'][5:]+' : public '+opaque['parent'][5:]+' {};\n')
            else:
                f.write('struct '+opaque['name'][5:]+' {};\n')

        f.write('} // namespace api\n')
        f.write('} // namespace anari\n')

        for opaque in anari['opaqueTypes']:
            f.write('typedef anari::api::'+opaque['name'][5:]+' *'+opaque['name']+';\n')

        f.write('#else\n')

        for opaque in anari['opaqueTypes']:
            f.write('typedef void* '+opaque['name']+';\n')

        f.write('#endif\n')

        f.write("""
#ifdef __cplusplus
extern "C" {
#endif

""")

        # generate structs
        for struct in anari['structs']:
            f.write('typedef struct {\n')

            if struct['members']:
                for member in struct['members']:
                    if 'elements' in member:
                        f.write('  '+member['type']+' '+member['name']+'['+str(member['elements'])+'];\n')
                    else:
                        f.write('  '+member['type']+' '+member['name']+';\n')

            f.write('} '+struct['name']+';\n')

        f.write('\n')
        # generate function prototypes
        for fun in anari['functionTypedefs']:
            f.write('typedef '+fun['returnType']+' (*'+fun['name']+')(')

            if fun['arguments']:
                arg = fun['arguments'][0]
                f.write(arg['type']+' '+arg['name'])
                for arg in fun['arguments'][1:]:
                    if 'default' in arg:
                        f.write(', '+arg['type']+' '+arg['name']+' ANARI_DEFAULT_VAL('+arg['default']+')')
                    else:
                        f.write(', '+arg['type']+' '+arg['name'])

            f.write(');\n')

        f.write('\n')
        # generate function prototypes
        for fun in anari['functions']:
            f.write('ANARI_INTERFACE '+fun['returnType']+' '+fun['name']+'(')

            if fun['arguments']:
                arg = fun['arguments'][0]
                f.write(arg['type']+' '+arg['name'])
                for arg in fun['arguments'][1:]:
                    if 'default' in arg:
                        f.write(', '+arg['type']+' '+arg['name']+' ANARI_DEFAULT_VAL('+arg['default']+')')
                    else:
                        f.write(', '+arg['type']+' '+arg['name'])

            f.write(');\n')

        f.write("""
#ifdef __cplusplus
} // extern "C"
#endif
""")


def write_anari_type_query_helper(filename, anari):
    os.makedirs(os.path.dirname(filename), exist_ok=True)

    with open(filename, mode='w') as f:
        f.write("// Copyright 2021-2026 The Khronos Group\n")
        f.write("// SPDX-License-Identifier: Apache-2.0\n\n")
        f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
        f.write("// Don't make changes to this directly\n\n")

        f.write("""
#pragma once

#include <anari/anari.h>
#include <math.h>
#include <stdint.h>

inline float anari_unit_clamp(float x) {
    if(x < -1.0f) {
        return -1.0f;
    } else if(x > 1.0f) {
        return 1.0f;
    } else {
        return x;
    }
}

inline int64_t anari_fixed_clamp(float x, int64_t max) {
    if(x <= -1.0f) {
        return -max;
    } else if(x >= 1.0f) {
        return max;
    } else {
        return (int64_t)(x*max);
    }
}

inline uint64_t anari_ufixed_clamp(float x, uint64_t max) {
    if(x <= 0.0f) {
        return 0u;
    } else if(x >= 1.0f) {
        return max;
    } else {
        return (uint64_t)(x*max);
    }
}

inline float anari_from_srgb(uint8_t x0) {
    float x = x0/(float)UINT8_MAX;
    if(x<=0.04045f) {
        return x*0.0773993808f;
    } else {
        return powf((x+0.055f)*0.94786729857f, 2.4f);
    }
}

inline uint8_t anari_to_srgb(float x) {
    if(x >= 1.0f) {
        return UINT8_MAX;
    } else if(x<=0.0f) {
        return 0u;
    } else if(x<=0.0031308f) {
        return (uint8_t)(x*12.92f*UINT8_MAX);
    } else {
        return (uint8_t)((powf(x*1.055f, 1.0f/2.4f)-0.055f)*UINT8_MAX);
    }
}


#ifdef __cplusplus

#include <utility>

namespace anari {

template<int type>
struct ANARITypeProperties { };

""")

        special_conversions = {}
        special_conversions['ANARI_UFIXED8_RGBA_SRGB'] = ("""        dst[0] = anari_from_srgb(src[0]);
        dst[1] = anari_from_srgb(src[1]);
        dst[2] = anari_from_srgb(src[2]);
        dst[3] = anari_unit_clamp(src[3]);
""", """        dst[0] = anari_to_srgb(src[0]);
        dst[1] = anari_to_srgb(src[1]);
        dst[2] = anari_to_srgb(src[2]);
        dst[3] = (base_type)anari_ufixed_clamp(float(src[3]), UINT8_MAX);
""")

        special_conversions['ANARI_UFIXED8_RGB_SRGB'] = ("""        dst[0] = anari_from_srgb(src[0]);
        dst[1] = anari_from_srgb(src[1]);
        dst[2] = anari_from_srgb(src[2]);
        dst[3] = 1.0f;
""", """        dst[0] = anari_to_srgb(src[0]);
        dst[1] = anari_to_srgb(src[1]);
        dst[2] = anari_to_srgb(src[2]);
""")

        special_conversions['ANARI_UFIXED8_RA_SRGB'] = ("""        dst[0] = anari_from_srgb(src[0]);
        dst[1] = 0;
        dst[2] = 0;
        dst[3] = anari_unit_clamp(src[1]/(float)UINT8_MAX);
""", """        dst[0] = anari_to_srgb(src[0]);
        dst[1] = (base_type)anari_ufixed_clamp(src[3], UINT8_MAX);
""")

        special_conversions['ANARI_UFIXED8_R_SRGB'] = ("""        dst[0] = anari_from_srgb(src[0]);
        dst[1] = 0.0f;
        dst[2] = 0.0f;
        dst[3] = 1.0f;
""", """        dst[0] = anari_to_srgb(src[0]);
""")

        enums = next(x for x in anari['enums'] if x['name']=='ANARIDataType')
        for enum in enums['values']:
            f.write('template<>\n')
            f.write('struct ANARITypeProperties<'+enum['name']+'> {\n')
            f.write('    using base_type = '+enum['baseType']+';\n')
            f.write('    static const int components = '+str(enum['elements'])+';\n')
            f.write('    static const bool normalized = '+('true' if enum['normalized'] else 'false')+';\n')
            f.write('    using array_type = base_type['+str(enum['elements'])+'];\n')
            f.write('    static constexpr const char* enum_name = "'+enum['name']+'";\n')
            f.write('    static constexpr const char* type_name = "'+enum['baseType']+'";\n')
            f.write('    static constexpr const char* array_name = "'+enum['baseType']+'['+str(enum['elements'])+']";\n')
            f.write('    static constexpr const char* var_name = "var'+enum['name'][6:].lower()+'";\n')

            if enum['value'] >= 1000: # conversions for numerical types
                f.write('    static void toFloat4(float *dst, const base_type *src) {\n')
                if enum['name'] in special_conversions:
                    f.write(special_conversions[enum['name']][0])
                else:
                    for i in range(0, 4):
                        if i < enum['elements']:
                            if 'FIXED' in enum['name']:
                                f.write('        dst['+str(i)+'] = anari_unit_clamp((float)src['+str(i)+']/(float)'+enum['baseType'].upper()[0:-2]+'_MAX);\n')
                            else:
                                f.write('        dst['+str(i)+'] = (float)src['+str(i)+'];\n')
                        else:
                            f.write('        dst['+str(i)+'] = '+('1.0f' if i==3 else '0.0f')+';\n')
                f.write('    }\n')

                f.write('    static void fromFloat4(base_type *dst, const float *src) {\n')
                if enum['name'] in special_conversions:
                    f.write(special_conversions[enum['name']][1])
                else:
                    for i in range(0, min(4, enum['elements'])):
                        if 'UFIXED' in enum['name']:
                            f.write('        dst['+str(i)+'] = (base_type)anari_ufixed_clamp(src['+str(i)+'], '+enum['baseType'].upper()[0:-2]+'_MAX);\n')
                        elif 'FIXED' in enum['name']:
                            f.write('        dst['+str(i)+'] = (base_type)anari_fixed_clamp(src['+str(i)+'], '+enum['baseType'].upper()[0:-2]+'_MAX);\n')
                        else:
                            f.write('        dst['+str(i)+'] = (base_type)src['+str(i)+'];\n')
                f.write('    }\n')
            else: # no op for non numeric types
                f.write('    static void fromFloat4(base_type *dst, const float *src) { }\n')
                f.write('    static void toFloat4(float *dst, const base_type *src) { }\n')



            f.write('};\n')

        f.write("""
template <typename R, template<int> class F, typename... Args>
R anariTypeInvoke(ANARIDataType type, Args&&... args) {
    switch (type) {
""")
        for enum in enums['values']:
            f.write('        case '+str(enum['value'])+': return F<'+str(enum['value'])+'>()(std::forward<Args>(args)...);\n')
        f.write('        default: return F<ANARI_UNKNOWN>()(std::forward<Args>(args)...);\n')
        f.write('    }\n')
        f.write('}\n')

        f.write('} // namespace anari\n')

        f.write('#endif\n')



        f.write("""
inline size_t anariSizeOf(ANARIDataType type) {
    switch (type) {
""")
        for enum in enums['values']:
            f.write('        case '+enum['name']+': return sizeof('+enum['baseType']+')*'+str(enum['elements'])+';\n')
        f.write('        default: return 4;\n')
        f.write('    }\n')
        f.write('}\n')



        f.write("""
inline size_t anariComponentsOf(ANARIDataType type) {
    switch (type) {
""")
        for enum in enums['values']:
            f.write('        case '+enum['name']+': return '+str(enum['elements'])+';\n')
        f.write('        default: return 1;\n')
        f.write('    }\n')
        f.write('}\n')



        f.write("""
inline const char* anariToString(ANARIDataType type) {
    switch (type) {
""")
        for enum in enums['values']:
            f.write('        case '+enum['name']+': return "'+enum['name']+'";\n')
        f.write('        default: return "ANARI_UNKNOWN";\n')
        f.write('    }\n')
        f.write('}\n')


        f.write("""
inline const char* anariTypenameOf(ANARIDataType type) {
    switch (type) {
""")
        for enum in enums['values']:
            f.write('        case '+enum['name']+': return "'+enum['baseType']+'";\n')
        f.write('        default: return "ANARI_UNKNOWN";\n')
        f.write('    }\n')
        f.write('}\n')


        f.write("""
inline const char* anariVarnameOf(ANARIDataType type) {
    switch (type) {
""")
        for enum in enums['values']:
            f.write('        case '+enum['name']+': return "var'+enum['name'][6:].lower()+'";\n')
        f.write('        default: return "ANARI_UNKNOWN";\n')
        f.write('    }\n')
        f.write('}\n')


        f.write("""
inline int anariIsNormalized(ANARIDataType type) {
    switch (type) {
""")
        for enum in enums['values']:
            f.write('        case '+enum['name']+': return '+str(int(enum['normalized']))+';\n')
        f.write('        default: return 0;\n')
        f.write('    }\n')
        f.write('}\n\n')

        f.write("""
#ifdef __cplusplus
namespace anari {

inline size_t sizeOf(ANARIDataType type)
{
    return anariSizeOf(type);
}

inline size_t componentsOf(ANARIDataType type)
{
    return anariComponentsOf(type);
}

inline const char* toString(ANARIDataType type)
{
    return anariToString(type);
}

inline const char* typenameOf(ANARIDataType type)
{
    return anariTypenameOf(type);
}

inline const char* varnameOf(ANARIDataType type)
{
    return anariVarnameOf(type);
}

inline int isNormalized(ANARIDataType type)
{
    return anariIsNormalized(type);
}

} // namespace anari
#endif
""")

output = sys.argv[1]
trees = [json.load(open(x)) for x in sys.argv[2:]]
api = trees[0]
for x in trees[1:]:
    merge_anari.merge(api, x)

filename = os.path.basename(output)
if filename == 'anari.h':
    write_anari_header(output, api)
elif filename == 'anari_enums.h':
    write_anari_enums(output, api)
elif filename == 'type_utility.h':
    write_anari_type_query_helper(output, api)



# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

import sys
import os
import json
import hash_gen
import merge_anari
import argparse
import pathlib
import subprocess

extra_params = {
    "ANARI_ARRAY1D" : [("const void*","appMemory"), ("ANARIMemoryDeleter","deleter"), ("const void*","userdata"), ("ANARIDataType","type"), ("uint64_t", "numItems1")],
    "ANARI_ARRAY2D" : [("const void*","appMemory"), ("ANARIMemoryDeleter","deleter"), ("const void*","userdata"), ("ANARIDataType","type"), ("uint64_t", "numItems1"), ("uint64_t", "numItems2")],
    "ANARI_ARRAY3D" : [("const void*","appMemory"), ("ANARIMemoryDeleter","deleter"), ("const void*","userdata"), ("ANARIDataType","type"), ("uint64_t", "numItems1"), ("uint64_t", "numItems2"), ("uint64_t", "numItems3")]
}

class FrontendGenerator:
    def __init__(self, anari, args):
        self.anari = anari
        self.args = args

        self.attribute_strings = {x["name"] for x in anari["attributes"]}
        self.enum_strings = set()
        self.param_strings = set()
        self.objects = dict()
        for obj in anari["objects"]:
            if not obj["type"] in self.objects:
                self.objects[obj["type"]] = list()
            if "name" in obj:
                self.objects[obj["type"]].append(obj["name"])


            self.param_strings.update([p["name"] for p in obj["parameters"]])
            for param in obj["parameters"]:
                if "ANARI_STRING" in param["types"] and "values" in param:
                    self.enum_strings.update(param["values"])

        self.all_strings = sorted(list(self.attribute_strings.union(self.enum_strings)))
        self.param_strings = sorted(list(self.param_strings))

        type_enums = next(x for x in anari["enums"] if x["name"] == "ANARIDataType")
        self.type_enum_dict = {e["name"]: e for e in type_enums["values"]}

    def hash_function(self):
        return hash_gen.gen_hash_function("parameter_string_hash", self.all_strings)

    def string_list(self):
        return "const char *param_strings[] = {\n   \"" + "\",\n   \"".join(self.all_strings) + "\"\n};\n"

    def define_list(self):
        code = "#define STRING_ENUM_unknown -1\n"
        code += "\n".join(["#define STRING_ENUM_%s %d"%(x, i) for i,x in enumerate(self.all_strings)])
        code += "\n"
        return code

    def format_as(self, value, anari_type):
        basetype = self.type_enum_dict[anari_type]["baseType"]
        if basetype == "float":
            return "%ff"%value
        elif basetype == "double":
            return "%f"%value
        elif basetype.startswith("uint") or basetype.startswith("int"):
            macro = basetype.upper()[:-1]+'C'
            return "%s(%d)"%(macro,value)
        else:
            return str(x)

    def declare_objects(self):
        code = ""
        code += "class ParameterPack {\n"
        code += "public:\n"
        code += "   virtual bool set(const char *paramname, ANARIDataType type, const void *mem) = 0;\n"
        code += "   virtual void unset(const char *paramname) = 0;\n"
        code += "   virtual ParameterBase& operator[](size_t idx) = 0;\n"
        code += "   virtual ParameterBase& operator[](const char *paramname) = 0;\n"
        code += "   virtual const char** paramNames() const = 0;\n"
        code += "   virtual size_t paramCount() const = 0;\n"
        code += "};\n"
        for idx, obj in enumerate(self.anari["objects"]):
            classname = obj["type"][6:].title()
            if "name" in obj:
                classname += obj["name"][0].upper() + obj["name"][1:]

            code += "class " + classname + " : public ParameterPack {\n"
            code += "public:\n"

            code += "   ANARIDevice device;\n"
            code += "   ANARIObject object;\n"

            code += "   static const int type = " + obj["type"] + ";\n"

            if "name" in obj:
                code += "   static constexpr const char *subtype = \"" + obj["name"] + "\";\n"
            else:
                code += "   static constexpr const char *subtype = nullptr;\n"

            code += "   static const uint32_t id = %d;\n"%idx

            for param in obj["parameters"]:
                code += "   Parameter<" + ", ".join(param["types"]) + "> " + param["name"].replace(".", "_") + ";\n"
            code += "\n"

            code += "   " + classname + "(ANARIDevice d, ANARIObject o);\n"

            code += "   bool set(const char *paramname, ANARIDataType type, const void *mem) override;\n"
            code += "   void unset(const char *paramname) override;\n"
            code += "   ParameterBase& operator[](size_t idx) override;\n"
            code += "   ParameterBase& operator[](const char *paramname) override;\n"
            code += "   const char** paramNames() const override;\n"
            code += "   size_t paramCount() const override;\n"

            code += "};\n"
        return code


    def set_default(self, param, indentation):
        code = ""
        paramname = param["name"].replace('.', '_')
        basetype = self.type_enum_dict[param["types"][0]]["baseType"]
        anari_type = param["types"][0]
        if isinstance(param["default"], list):
            code += indentation + basetype + " value[] = {" + ", ".join([self.format_as(x, anari_type) for x in param["default"]]) + "};\n"
        elif isinstance(param["default"], str):
            code += indentation + "const char *value = \"" + param["default"] + "\";\n"
        else:
            code += indentation + basetype + " value[] = {" + self.format_as(param["default"], anari_type) + "};\n"
        code += indentation + "%s.set(device, object, %s, value);\n"%(paramname, param["types"][0])
        return code

    def implement_objects(self):
        code = ""
        code += "static " + hash_gen.gen_hash_function("param_hash", self.param_strings, indent = "")
        for obj in self.anari["objects"]:
            classname = obj["type"][6:].title()
            if "name" in obj:
                classname += obj["name"][0].upper() + obj["name"][1:]

            code += classname + "::" + classname + "(ANARIDevice device, ANARIObject o) : device(device), object(o)"
            code += " {\n"

            for param in obj["parameters"]:
                if "default" in param:
                    code += "   {\n"
                    code += self.set_default(param, indentation="      ")
                    code += "   }\n"

            code += "}\n"
            code += "bool " + classname + "::set(const char *paramname, ANARIDataType type, const void *mem) {\n"
            code += "   int idx = param_hash(paramname);\n"
            code += "   switch(idx) {\n"
            for j,param in enumerate(obj["parameters"]):
                i = self.param_strings.index(param["name"])
                paramname = param["name"].replace('.', '_')
                code += "      case %d: //%s\n"%(i,param["name"])
                code += "         return %s.set(device, object, type, mem);\n"%paramname
            code += "      default: // unknown param\n"
            code += "         //unknown parameter\n"
            code += "         return false;\n"
            code += "   }\n"
            code += "}\n"
            code += "void " + classname + "::unset(const char *paramname) {\n"
            code += "   int idx = param_hash(paramname);\n"
            code += "   switch(idx) {\n"
            for j,param in enumerate(obj["parameters"]):
                i = self.param_strings.index(param["name"])
                paramname = param["name"].replace('.', '_')
                code += "      case %d: //%s\n"%(i,param["name"])
                if "default" in param:
                    code += "         {\n"
                    code += self.set_default(param, indentation="            ")
                    code += "         }\n"
                else:
                    code += "         %s.unset(device, object);\n"%paramname
                code += "         return;\n"
            code += "      default: // unknown param\n"
            code += "         //unknown parameter\n"
            code += "         return;\n"
            code += "   }\n"
            code += "}\n"


            code += "ParameterBase& " + classname + "::operator[](size_t idx) {\n"
            code += "   static EmptyParameter empty;\n"
            code += "   switch(idx) {\n"
            for idx, param in enumerate(obj["parameters"]):
                code += "      case %d: return "%idx + param["name"].replace(".", "_") + ";\n"
            code += "      default: return empty;\n"
            code += "   }\n"
            code += "}\n"

            code += "ParameterBase& " + classname + "::operator[](const char *paramname) {\n"
            code += "   static EmptyParameter empty;\n"
            code += "   int idx = param_hash(paramname);\n"
            code += "   switch(idx) {\n"
            for param in obj["parameters"]:
                idx = self.param_strings.index(param["name"])
                code += "      case %d: return "%idx + param["name"].replace(".", "_") + ";\n"
            code += "      default: return empty;\n"
            code += "   }\n"
            code += "}\n"

            code += "const char ** " + classname + "::paramNames() const {\n"

            code += "   static const char *paramnames[] = {\n"
            for param in obj["parameters"]:
                code += "      \"" + param["name"] + "\",\n"
            code += "      nullptr\n"
            code += "   };\n"
            code += "   return paramnames;\n"
            code += "}\n"

            code += "size_t " + classname + "::paramCount() const {\n"
            code += "   return %d;\n"%len(obj["parameters"])
            code += "}\n"

            code += "\n"
        return code

    def implement_device(self):
        code = ""
        objnames = set()
        for key, names in self.objects.items():
            objnames.update(names)
        objnames = sorted(list(objnames))

        code += "static " + hash_gen.gen_hash_function("obj_hash", objnames, indent = "")

        devicename = self.args.prefix + "Device"
        for key, names in self.objects.items():
            if key == "ANARI_DEVICE":
                continue
            factoryname = "".join(key.title().split("_")[1:])

            if names:
                code += "ANARI" + factoryname + " "
                code += devicename + "::new" + factoryname + "(const char *type) {\n"
                code += "   int idx = obj_hash(type);\n"
                code += "   switch(idx) {\n"
                for name in names:
                    classname = key[6:].title() + name[0].upper() + name[1:]

                    code += "      case %d: //%s\n"%(objnames.index(name), name)
                    code += "         return allocate<ANARI"+factoryname+", "+classname+">();\n"

                code += "      default: // unknown object\n"
                code += "         return 0;\n"
                code += "   }\n"
                code += "   return 0;\n"
                code += "}\n"
            else:
                code += "ANARI" + factoryname + " "
                code += devicename + "::new" + factoryname + "("
                if key in extra_params:
                    code += ", ".join("%s %s"%x for x in extra_params[key])
                code += ") {\n"
                classname = key[6:].title()
                code += "   return allocate<ANARI"+factoryname+", "+classname+">("
                if key in extra_params:
                    code += ", ".join(x[1] for x in extra_params[key])
                code += ");\n"
                code += "}\n"
        return code

parser = argparse.ArgumentParser(description="Generate frontend for an ANARI device.")
parser.add_argument("-d", "--device", dest="devicespec", type=pathlib.Path, required=True, help="The device json file.")
parser.add_argument("-j", "--json", dest="json", required=True, type=pathlib.Path, action="append", help="Path to the core and extension json root.")
parser.add_argument("-p", "--prefix", dest="prefix", required=True, help="Prefix for the classes and filenames.")
parser.add_argument("--namespace", dest="namespace", required=True, help="Namespace for generated code. Also used as library name.")
parser.add_argument("--name", dest="name", required=True, help="Name of the library and device.")
parser.add_argument("-f", "--force", dest="force", action="store_true", help="Force regeneration of all files. This will overwrite edited files.")
parser.add_argument("-l", "--header", dest="header", help="File containing a header to be inserted at the top of source files")
parser.add_argument("-o", "--output", dest="outdir", type=pathlib.Path, default=pathlib.Path("."), help="Output directory")
parser.add_argument("--omit_target_prefix", dest="omitprefix", action="store_true", help=argparse.SUPPRESS)
args = parser.parse_args()

try:
    devicespec = args.devicespec.relative_to(args.outdir)
except ValueError as err:
    print("device specification needs to be in output directory.")
    print(err)
    sys.exit(1)

templatepath = pathlib.Path(__file__).parent / "templates"

#flattened list of all input jsons in supplied directories
jsons = [entry for j in args.json for entry in j.glob("**/*.json")]

#load the device root
device = json.load(open(args.devicespec))
print("opened " + device["info"]["type"] + " " + device["info"]["name"])

#load header if present
header = ""
if args.header:
    header = open(args.header).read()

dependencies = merge_anari.crawl_dependencies(device, jsons)
#merge all dependencies
for x in dependencies:
    matches = [p for p in jsons if p.stem == x]
    for m in matches:
        feature = json.load(open(m))
        merge_anari.assemble(feature, jsons)
        merge_anari.merge(device, feature)

#generate files
gen = FrontendGenerator(device, args)


rootdir = args.outdir
gendir = args.outdir/"generated"
srcdir = args.outdir/"src"

rootdir.mkdir(parents=True, exist_ok=True)
gendir.mkdir(parents=True, exist_ok=True)
srcdir.mkdir(parents=True, exist_ok=True)

targetprefix = "anari::"
if args.omitprefix:
    targetprefix = ""

namespaces = args.namespace.split("::")
begin_namespaces = "".join(["namespace %s{\n"%n for n in namespaces])
end_namespaces = "".join(["} //namespace %s\n"%n for n in reversed(namespaces)])



with open(gendir/(args.prefix + "String.h"), mode='w') as f:
    f.write(header)
    f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
    f.write("// Don't make changes to this directly\n\n")
    f.write("#pragma once\n")
    f.write(begin_namespaces)
    f.write(gen.define_list())
    f.write("extern const char *param_strings[];\n")
    f.write("int parameter_string_hash(const char *str);\n")
    f.write(end_namespaces)

with open(gendir/(args.prefix + "String.cpp"), mode='w') as f:
    f.write(header)
    f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
    f.write("// Don't make changes to this directly\n\n")
    f.write("#include \""+args.prefix+"String.h\"\n")
    f.write("#include <cstdint>\n")
    f.write(begin_namespaces)
    f.write(gen.hash_function())
    f.write(gen.string_list())
    f.write(end_namespaces)

with open(gendir/(args.prefix + "Objects.h"), mode='w') as f:
    f.write(header)
    f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
    f.write("// Don't make changes to this directly\n\n")
    f.write("#pragma once\n")
    f.write("#include \""+args.prefix+"Parameter.h\"\n")
    f.write(begin_namespaces)
    f.write(gen.declare_objects())
    f.write(end_namespaces)

with open(gendir/(args.prefix + "Objects.cpp"), mode='w') as f:
    f.write(header)
    f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
    f.write("// Don't make changes to this directly\n\n")

    f.write("#include <stdint.h>\n")
    f.write("#include \""+args.prefix+"Objects.h\"\n")
    f.write(begin_namespaces)
    f.write(gen.implement_objects())
    f.write(end_namespaces)

with open(gendir/(args.prefix + "DeviceFactories.cpp"), mode='w') as f:
    f.write(header)
    f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
    f.write("// Don't make changes to this directly\n\n")

    f.write("#include \""+args.prefix+"Device.h\"\n")
    f.write("#include \""+args.prefix+"Objects.h\"\n")
    f.write("#include \""+args.prefix+"Specializations.h\"\n")
    f.write(begin_namespaces)
    f.write(gen.implement_device())
    f.write(end_namespaces)

def generate_if_not_present(targetdir, name, header, noprefix=False):
    filename = args.prefix + name
    if noprefix:
        filename = name
    filename = targetdir/filename
    if args.force and filename.is_file():
        os.rename(filename, filename.parent/(filename.name + ".old"))
    if not filename.is_file():
        with open(filename, mode='w') as f:
            infile = (templatepath/name).open().read()
            infile = infile.replace("$libraryname", args.name)
            infile = infile.replace("$namespace", args.namespace)
            infile = infile.replace("$begin_namespaces", begin_namespaces)
            infile = infile.replace("$end_namespaces", end_namespaces)
            infile = infile.replace("$prefix", args.prefix)
            infile = infile.replace("$generator", os.path.basename(__file__))
            infile = infile.replace("$device_json", str(args.devicespec.relative_to(rootdir)))
            infile = infile.replace("$device_json", "foo")
            infile = infile.replace("$generated_path", str(gendir.relative_to(rootdir)))
            infile = infile.replace("$source_path", str(srcdir.relative_to(rootdir)))
            infile = infile.replace("$target_prefix", targetprefix)
            infile = infile.replace("$template", name)
            f.write(header)
            f.write(infile)

generate_if_not_present(gendir, "Parameter.h", header)
generate_if_not_present(srcdir, "Object.h", header)
generate_if_not_present(srcdir, "ArrayObjects.h", header)
generate_if_not_present(srcdir, "ArrayObjects.cpp", header)
generate_if_not_present(srcdir, "FrameObject.h", header)
generate_if_not_present(srcdir, "FrameObject.cpp", header)
generate_if_not_present(srcdir, "DeviceObject.h", header)
generate_if_not_present(srcdir, "DeviceObject.cpp", header)
generate_if_not_present(srcdir, "Device.h", header)
generate_if_not_present(srcdir, "Device.cpp", header)
generate_if_not_present(srcdir, "Specializations.h", header)
generate_if_not_present(rootdir, "CMakeLists.txt", header.replace("//","#"), noprefix = True)


forwarded = ["--device", args.devicespec]
forwarded += ["--prefix", args.prefix]
forwarded += ["--namespace", args.namespace]
for x in args.json:
    forwarded += ["--json", str(x)]

scriptdir = pathlib.Path(__file__).parent
subprocess.run([sys.executable, scriptdir/"generate_queries.py","--output", str(gendir)]+forwarded)



# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

# this script checks if a device definition is complete

import sys
import os
import json
import hash_gen
import merge_anari
import argparse
import pathlib

def check_feature_completeness(feature, filename):
    if "info" not in feature:
        print("%s is missing info section"%filename)
        return

    info = feature["info"]
    if "name" not in info:
        print("%s info section is missing name"%filename)

    if "type" not in info:
        print("%s info section is missing type"%filename)


def check_object_completeness(device):
    for obj in device["objects"]:
        objectname = ""
        if "type" not in obj:
            if "name" in obj:
                print("object \"%s\" missing type"%obj["name"])
            else:
                print("anonymous object missing type")
        else:
            objectname += obj["type"]

        if "name" in obj:
            objectname += " \"%s\""%obj["name"]

        if "description" not in obj:
            print("%s missing description"%objectname)

        if "parameters" not in obj:
            print("%s: missing parameters"%objectname)
        else:
            for param in obj["parameters"]:
                if "name" not in param:
                    print("%s: parameter missing name"%objectname)

                paramname = param["name"]

                if "types" not in param:
                    print("%s parameter \"%s\" missing types"%(objectname, paramname))
                elif not isinstance(param["types"], list):
                    print("%s parameter \"%s\" types must be array"%(objectname, paramname))
                elif not param["types"]:
                    print("%s parameter \"%s\" types list can't be empty"%(objectname, paramname))

                if "ANARI_ARRAY1D" in param["types"] or "ANARI_ARRAY2D" in param["types"] or "ANARI_ARRAY3D" in param["types"]:
                    if "elementType" not in param:
                        print("%s parameter \"%s\" allows arrays but is missing elementType"%(objectname, paramname))
                    elif not isinstance(param["elementType"], list):
                        print("%s parameter \"%s\" elementType must be an array"%(objectname, paramname))
                    elif not param["elementType"]:
                        print("%s parameter \"%s\" elementType list can't be empty"%(objectname, paramname))

                if "tags" not in param:
                    print("%s parameter \"%s\"  missing tags"%(objectname, paramname))
                elif not isinstance(param["types"], list):
                    print("%s parameter \"%s\" tags must be an array"%(objectname, paramname))

                if "description" not in param:
                    print("%s parameter \"%s\" missing description"%(objectname, paramname))


parser = argparse.ArgumentParser(description="Generate query functions for an ANARI device.")
parser.add_argument("-d", "--device", dest="devicespec", type=open, help="The device json file.")
parser.add_argument("-j", "--json", dest="json", type=pathlib.Path, action="append", help="Path to the core and extension json root.")
args = parser.parse_args()

#flattened list of all input jsons in supplied directories
jsons = [entry for j in args.json for entry in j.glob("**/*.json")]

#load the device root
device = json.load(args.devicespec)
merge_anari.tag_extension(device)
print("opened " + device["info"]["type"] + " " + device["info"]["name"])

dependencies = merge_anari.crawl_dependencies(device, jsons)
#merge all dependencies
for x in dependencies:
    matches = [p for p in jsons if p.stem == x]
    for m in matches:
        feature = json.load(open(m))
        check_feature_completeness(feature, m)
        merge_anari.tag_extension(feature)
        merge_anari.merge(device, feature)

check_object_completeness(device)


import sys
import os
import json

def check_conflicts(a, b, key, scope):
    a_set = set([x[key] for x in a])
    b_set = set([x[key] for x in b])
    conflicts = set.intersection(a_set, b_set)
    for conflict in conflicts:
        entries = [x for x in a if x[key] == conflict] + [x for x in b if x[key] == conflict]
        print('duplicate '+key+' while merging '+scope)
        for e in entries:
            print(e)

def merge_enums(core, extension):
    for enum in extension:
        name = enum['name']
        c = next(x for x in core if x['name'] == name)
        if c:
            check_conflicts(c['values'], enum['values'], 'name', name)
            check_conflicts(c['values'], enum['values'], 'value', name)
            c['values'].extend(enum['values'])
        else:
            core.extend(extension)

def merge(core, extension):
    for k in extension.keys():
        if k == "enums" :
            merge_enums(core[k], extension[k])
        elif k == "info":
            print('merging '+extension[k]['type']+' '+extension[k]["name"])
        elif k == "descriptions":
            core.extend(extension)
        else:
            check_conflicts(core[k], extension[k], 'name', k)
            core[k].extend(extension[k])



def write_anari_enum_list(anari):
    f = open("enums.txt", mode='w')
    # generate enums
    for enum in anari['enums']:
        for value in enum['values']:
            f.write(value['name']+'\n')

def write_anari_function_list(anari):
    f = open("functions.txt", mode='w')
    # generate enums
    for func in anari['functions']:
        f.write(func['name']+'\n')

def write_anari_type_list(anari):
    f = open("types.txt", mode='w')
    # generate enums
    for x in anari['enums']:
        f.write(x['name']+'\n')
    for x in anari['opaqueTypes']:
        f.write(x['name']+'\n')
    for x in anari['functionTypedefs']:
        f.write(x['name']+'\n')
    for x in anari['structs']:
        f.write(x['name']+'\n')

trees = [json.load(open(x)) for x in sys.argv[1:]]
api = trees[0]
for x in trees[1:]:
    merge(api, x)


write_anari_enum_list(api)
write_anari_function_list(api)
write_anari_type_list(api)


# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

import sys
import os
import json
import hash_gen
import merge_anari
import argparse
import pathlib

"""static void fillStruct(ANARIExtensions *extensions, const char **list) {

    for(const char *i = *list;i!=NULL;++i) {
        switch(extension_hash(i)) {

        }
    }
}

"""

class ExtensionUtilityGenerator:
    def __init__(self, anari, extension_names):
        self.anari = anari
        self.extensions = ["ANARI_"+f for f in self.anari["extensions"]]
        self.extension_names = extension_names

    def generate_extension_header(self):
        code = "typedef struct {\n"
        for extension in self.extensions:
            code += "   int %s;\n"%extension
        code += "} ANARIExtensions;\n"
        code += "// Return all extension identifiers defined by the API JSON inputs.\n"
        code += "// The returned array is terminated by a null pointer.\n"
        code += "static inline const char *const *anariGetExtensionNames(void) {\n"
        code += "    static const char *const names[] = {\n"
        for extension in self.extension_names:
            code += "        \"%s\",\n"%extension
        code += "        0\n"
        code += "    };\n"
        code += "    return names;\n"
        code += "}\n"
        code += "int anariGetDeviceExtensionStruct(ANARIExtensions *extensions, ANARILibrary library, const char *deviceName);\n"
        code += "int anariGetObjectExtensionStruct(ANARIExtensions *extensions, ANARIDevice device, ANARIDataType objectType, const char *objectName);\n"
        code += "int anariGetInstanceExtensionStruct(ANARIExtensions *extensions, ANARIDevice device, ANARIObject object);\n"

        code += "#ifdef ANARI_EXTENSION_UTILITY_IMPL\n"
        code += "#include <string.h>\n"
        code += "static " + hash_gen.gen_hash_function("extension_hash", self.extensions)
        code += "static void fillExtensionStruct(ANARIExtensions *extensions, const char *const *list) {\n"
        code += "    memset(extensions, 0, sizeof(ANARIExtensions));\n"
        code += "    for(const char *const *i = list;*i!=NULL;++i) {\n"
        code += "        switch(extension_hash(*i)) {\n"
        for idx, extension in enumerate(self.extensions):
            code += "            case %d: extensions->%s = 1; break;\n"%(idx, extension)
        code += "            default: break;\n"
        code += "        }\n"
        code += "    }\n"
        code += "}\n"
        code += "int anariGetDeviceExtensionStruct(ANARIExtensions *extensions, ANARILibrary library, const char *deviceName) {\n"
        code += "    const char *const *list = (const char *const *)anariGetDeviceExtensions(library, deviceName);\n"
        code += "    if(list) {\n"
        code += "        fillExtensionStruct(extensions, list);\n"
        code += "        return 0;\n"
        code += "    } else {\n"
        code += "        return 1;\n"
        code += "    }\n"
        code += "}\n"
        code += "int anariGetObjectExtensionStruct(ANARIExtensions *extensions, ANARIDevice device, ANARIDataType objectType, const char *objectName) {\n"
        code += "    const char *const *list = (const char *const *)anariGetObjectInfo(device, objectType, objectName, \"extension\", ANARI_STRING_LIST);\n"
        code += "    if(list) {\n"
        code += "        fillExtensionStruct(extensions, list);\n"
        code += "        return 0;\n"
        code += "    } else {\n"
        code += "        return 1;\n"
        code += "    }\n"
        code += "}\n"
        code += "int anariGetInstanceExtensionStruct(ANARIExtensions *extensions, ANARIDevice device, ANARIObject object) {\n"
        code += "    const char *const *list = NULL;\n"
        code += "    anariGetProperty(device, object, \"extension\", ANARI_STRING_LIST, &list, sizeof(list), ANARI_WAIT);\n"
        code += "    if(list) {\n"
        code += "        fillExtensionStruct(extensions, list);\n"
        code += "        return 0;\n"
        code += "    } else {\n"
        code += "        return 1;\n"
        code += "    }\n"
        code += "}\n"
        code += "#endif\n"
        return code


parser = argparse.ArgumentParser(description="Generate query functions for an ANARI device.")
parser.add_argument("-d", "--device", dest="devicespec", type=open, help="The device json file.")
parser.add_argument("-j", "--json", dest="json", type=pathlib.Path, action="append", help="Path to the core and extension json root.")
parser.add_argument("-n", "--namespace", dest="namespace", help="Namespace for the classes and filenames.")
parser.add_argument("-o", "--output", dest="outdir", type=pathlib.Path, default=pathlib.Path("."), help="Output directory")
args = parser.parse_args()


#flattened list of all input jsons in supplied directories
jsons = [entry for j in args.json for entry in j.glob("**/*.json")]

#load the device root
device = json.load(args.devicespec)
merge_anari.tag_extension(device)
print("opened " + device["info"]["type"] + " " + device["info"]["name"])

dependencies = merge_anari.crawl_dependencies(device, jsons)
#merge all dependencies
for x in dependencies:
    matches = [p for p in jsons if p.stem == x]
    for m in matches:
        extension = json.load(open(m))
        merge_anari.tag_extension(extension)
        merge_anari.merge(device, extension)

def canonical_extension_names(json_paths):
    names = set()
    for path in json_paths:
        with open(path) as source:
            definition = json.load(source)
        if definition.get("info", {}).get("type") == "extension":
            names.add("ANARI_" + definition["info"]["name"])
    return sorted(names)


#generate files
gen = ExtensionUtilityGenerator(device, canonical_extension_names(jsons))


def begin_namespaces(args):
    output = ""
    if args.namespace:
        for n in args.namespace.split("::"):
            output += "namespace %s {\n"%n
    return output

def end_namespaces(args):
    output = ""
    if args.namespace:
        for n in args.namespace.split("::"):
            output += "}\n"
    return output


with open(args.outdir/"anari_extension_utility.h", mode='w') as f:
    f.write("// Copyright 2021-2026 The Khronos Group\n")
    f.write("// SPDX-License-Identifier: Apache-2.0\n\n")
    f.write("// This file was generated by "+os.path.basename(__file__)+"\n")
    f.write("// Don't make changes to this directly\n\n")

    f.write("#pragma once\n")
    f.write("#include <anari/anari.h>\n")
    f.write(gen.generate_extension_header())



# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

import sys
import os
import json

def gen_hash_function_table_recursive(table, i, current, keywords, values):
   offset = len(table)
   crange = [ord(x[i]) for x in current]
   low = min(crange)
   high = max(crange)+1
   table.extend([0]*(high-low))
   for c in range(low, high):
      s = [x for x in current if i<len(x) and ord(x[i])==c]
      if s:
         if len(s) == 1 and c == 0:
            idx = keywords.index(s[0][0:-1])
            table[offset+c-low] = values[idx]+(0x80000000)
         else:
            table[offset+c-low] = gen_hash_function_table_recursive(table, i+1, s, keywords, values)

   return offset + (low<<16) + (high<<24)



def gen_hash_function_table(keywords, values):
   table = []
   root_keywords = [x+'\0' for x in keywords]
   root = gen_hash_function_table_recursive(table, 0, root_keywords, keywords, values)
   return (root, table)

def gen_hash_function(name, keywords, values=None, indent=""):
   if values==None:
      values = range(0, len(keywords))

   root, table = gen_hash_function_table(keywords, values)

   code  = ""
   code += "int %s(const char *str) {\n"%name
   code += indent + "   static const uint32_t table[] = {"
   code += ",".join(["0x%xu"%x for x in table])
   code += "};\n"
   code += indent + "   uint32_t cur = 0x%xu;\n"%root
   code += indent + "   for(int i = 0;cur!=0;++i) {\n"
   code += indent + "      uint32_t idx = cur&0xFFFFu;\n"
   code += indent + "      uint32_t low = (cur>>16u)&0xFFu;\n"
   code += indent + "      uint32_t high = (cur>>24u)&0xFFu;\n"
   code += indent + "      uint32_t c = (uint32_t)str[i];\n"
   code += indent + "      if(c>=low && c<high) {\n"
   code += indent + "         cur = table[idx+c-low];\n"
   code += indent + "      } else {\n"
   code += indent + "         break;\n"
   code += indent + "      }\n"
   code += indent + "      if(cur&0x80000000u) {\n"
   code += indent + "         return cur&0xFFFFu;\n"
   code += indent + "      }\n"
   code += indent + "      if(str[i]==0) {\n"
   code += indent + "         break;\n"
   code += indent + "      }\n"
   code += indent + "   }\n"
   code += indent + "   return -1;\n"
   code += indent + "}\n"
   return code

def array_of_strings(name, keywords, values=None):
   code  = ""
   code += "const char *%s_strings[] = {"%name
   code += ",".join(["\"%s\""%x for x in keywords])
   code += "};\n"
   return code

def define_enums(name, keywords, values=None):
   if values==None:
      values = range(0, len(keywords))
   code  = ""
   code += "\n".join(["#define %s_%s %d"%(name.upper(), x.upper(), values[keywords.index(x)]) for x in keywords])
   code += "\n"
   return code



# Copyright 2021-2026 The Khronos Group
# SPDX-License-Identifier: Apache-2.0

import json

def check_conflicts(a, b, key, scope):
    a_set = set([x[key] for x in a])
    b_set = set([x[key] for x in b])
    conflicts = set.intersection(a_set, b_set)
    for conflict in conflicts:
        entries = [x for x in a if x[key] == conflict] + [x for x in b if x[key] == conflict]
        print('duplicate '+key+' while merging '+scope)
        for e in entries:
            print(e)

def merge_enums(core, extension):
    for enum in extension:
        name = enum['name']
        c = next((x for x in core if x['name'] == name), None)
        if c:
            check_conflicts(c['values'], enum['values'], 'name', name)
            check_conflicts(c['values'], enum['values'], 'value', name)
            c['values'].extend(enum['values'])
        else:
            core.extend(extension)

def merge_api(core, extension):
    for k in extension.keys():
        if k == "enums" :
            merge_enums(core[k], extension[k])
        elif k == "info":
            print('merging '+extension[k]['type']+' '+extension[k]["name"])
        elif k == "descriptions":
            core.extend(extension)
        else:
            check_conflicts(core[k], extension[k], 'name', k)
            core[k].extend(extension[k])

def merge_parameters(core, extension):
    for param in extension:
        name = param["name"]
        c = next((x for x in core if x["name"] == name), None)
        if c:
            if "types" in c and "types" in param:
                c["types"].extend(param["types"])
            if "values" in c and "values" in param:
                c["values"].extend(param["values"])
        else:
            core.append(param)

def merge_objects(core, extension):
    for field in extension:
        if field == "parameters":
            merge_parameters(core["parameters"], extension["parameters"])
        elif field in core:
            if isinstance(core[field], list):
                core[field].extend(extension[field])
        elif field != "sourceExtension":
            core[field] = extension[field]

def merge_object_table(core, extension, wildcard = ""):
    for obj in extension:
        if "name" in obj and obj["name"] != wildcard:
            name = obj["name"]
            kind = obj["type"]
            c = next((x for x in core if x["type"] == kind and "name" in x and x["name"] == name), None)
            if c:
                merge_objects(c, obj)
            else:
                core.append(obj)
        else:
            kind = obj["type"]
            c = next((x for x in core if x["type"] == kind), None)
            if c:
                merge_objects(c, obj)
            else:
                core.append(obj)

def merge(core, extension, verbose=False):
    if not "extensions" in core:
        core["extensions"] = []
        if core["info"]["type"] == "extension" and not core["info"]["name"] in core["extensions"]:
                core["extensions"].append(core["info"]["name"])
    for k,v in extension.items():
        if not k in core:
            core[k] = v
        elif k == "info":
            if verbose:
                print('merging '+extension[k]['type']+' '+extension[k]["name"])
            if "name" in v and v["type"] == "extension" and not v["name"] in core["extensions"]:
                core["extensions"].append(v["name"])
        elif k == "enums" :
            merge_enums(core[k], extension[k])
        elif k == "descriptions":
            core.extend(extension)
        elif k == "objects":
            merge_object_table(core[k], extension[k])
        else:
            check_conflicts(core[k], extension[k], 'name', k)
            core[k].extend(extension[k])

def tag_extension(tree):
    if "info" in tree and "name" in tree["info"] and tree['info']['type'] == 'extension':
        extension = tree["info"]["name"]
        if "objects" in tree:
            for obj in tree["objects"]:
                obj["sourceExtension"] = extension
                if "parameters" in obj:
                    for param in obj["parameters"]:
                        param["sourceExtension"] = extension

def assemble(tree, jsons):
    if "implements" in tree["info"]:
        for extension in tree["info"]["implements"]:
            matches = [p for p in jsons if p.stem == extension]
            for m in matches:
                ext = json.load(open(m))
                merge_object_table(tree["objects"], ext["objects"], "default")

def crawl_dependencies(root, jsons):
    deps = []
    deps.extend(reversed(root["info"]["dependencies"].copy()))
    i = 0
    while i<len(deps):
        x = deps[i]
        i += 1
        matches = [p for p in jsons if p.stem == x]
        for m in matches:
            f = json.load(open(m))
            if "info" in f and "dependencies" in f["info"]:
                deps.extend(reversed(f["info"]["dependencies"]))
    duplicates = set()
    return [x for x in reversed(deps) if not (x in duplicates or duplicates.add(x))]